import base64
import contextlib
import datetime
import logging
from _ctypes import POINTER, byref
from ctypes import c_ubyte, c_uint32, c_uint8, c_int, c_char_p, cast, c_uint, c_void_p
from typing import List, BinaryIO

from . import _worm
from .errors import _guard
from .utils import log_execution

logger = logging.getLogger(__name__)


def get_version() -> str:
    """
    Returns the library version.
    """
    return _worm.worm_getVersion().decode()


def is_online_sdk() -> bool:
    """
    Returns whether this SDK is an online SDK or not.
    """
    return bool(_worm.worm_isOnlineSdk())


def signature_algorithm() -> str:
    """
    Returns the signature algorithm that is used by the TSE.
    """
    return _worm.worm_signatureAlgorithm().decode()


def log_time_format() -> str:
    """
    Returns the log time format used by the TSE.
    """
    return _worm.worm_logTimeFormat().decode()


def _c_ubyte(bs):
    return (c_ubyte * len(bs))(*bs)


class BaseWormContext:
    def __init__(self):
        self._ctx = None

    def close(self):
        _worm.worm_cleanup(self._ctx)
        del self._ctx

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def keepalive_configure(self, seconds: int):
        _guard(_worm.worm_keepalive_configure(self._ctx, seconds))

    def keepalive_disable(self):
        _guard(_worm.worm_keepalive_disable(self._ctx))

    @log_execution
    def info(self) -> dict:
        i = _worm.worm_info_new(self._ctx)
        _guard(_worm.worm_info_read(i))

        val_res = POINTER(c_ubyte)()
        val_len = _worm.worm_uint()
        _worm.worm_info_tsePublicKey(i, byref(val_res), byref(val_len))
        pubkey = bytes([val_res[i] for i in range(val_len.value)])

        val_res = POINTER(c_ubyte)()
        val_len = _worm.worm_uint()
        _worm.worm_info_tseSerialNumber(i, byref(val_res), byref(val_len))
        serial = bytes([val_res[i] for i in range(val_len.value)])
        states = {
            _worm.WORM_INIT_UNINITIALIZED: "WORM_INIT_UNINITIALIZED",
            _worm.WORM_INIT_INITIALIZED: "WORM_INIT_INITIALIZED",
            _worm.WORM_INIT_DECOMMISSIONED: "WORM_INIT_DECOMMISSIONED",
        }

        info = {
            "isDevelopmentFirmware": bool(_worm.worm_info_isDevelopmentFirmware(i)),
            "capacity": _worm.worm_info_capacity(i),
            "size": _worm.worm_info_size(i),
            "hasValidTime": bool(_worm.worm_info_hasValidTime(i)),
            "hasPassedSelfTest": bool(_worm.worm_info_hasPassedSelfTest(i)),
            "isCtssInterfaceActive": bool(_worm.worm_info_isCtssInterfaceActive(i)),
            "isExportEnabledIfCspTestFails": bool(
                _worm.worm_info_isExportEnabledIfCspTestFails(i)
            ),
            "initializationState": states[_worm.worm_info_initializationState(i)],
            "hasChangedPuk": bool(_worm.worm_info_hasChangedPuk(i)),
            "hasChangedAdminPin": bool(_worm.worm_info_hasChangedAdminPin(i)),
            "timeUntilNextSelfTest": _worm.worm_info_timeUntilNextSelfTest(i),
            "startedTransactions": _worm.worm_info_startedTransactions(i),
            "maxStartedTransactions": _worm.worm_info_maxStartedTransactions(i),
            "createdSignatures": _worm.worm_info_createdSignatures(i),
            "maxSignatures": _worm.worm_info_maxSignatures(i),
            "remainingSignatures": _worm.worm_info_remainingSignatures(i),
            "maxTimeSynchronizationDelay": _worm.worm_info_maxTimeSynchronizationDelay(
                i
            ),
            "maxUpdateDelay": _worm.worm_info_maxUpdateDelay(i),
            "tsePublicKey": pubkey,
            "timeUntilNextTimeSynchronization": _worm.worm_info_timeUntilNextTimeSynchronization(
                i
            ),
            "tseSerialNumberBytes": serial,
            "tseSerialNumberHex": serial.hex(),
            "tseDescription": _worm.worm_info_tseDescription(i).decode(),
            "registeredClients": _worm.worm_info_registeredClients(i),
            "maxRegisteredClients": _worm.worm_info_maxRegisteredClients(i),
            "certificateExpirationDate": datetime.datetime.fromtimestamp(
                _worm.worm_info_certificateExpirationDate(i), datetime.timezone.utc
            ),
            "tarExportSizeInSectors": _worm.worm_info_tarExportSizeInSectors(i),
            "tarExportSize": _worm.worm_info_tarExportSize(i),
            "hardwareVersion": _worm.worm_info_hardwareVersion(i),
            "softwareVersion": _worm.worm_info_softwareVersion(i),
            "formFactor": _worm.worm_info_formFactor(i).decode(),
        }
        _worm.worm_info_free(i)
        return info

    @log_execution
    def flash_health(self) -> dict:
        val_errors = c_uint32()
        val_spare = c_uint8()
        val_erase = c_uint8()
        val_retention = c_uint8()
        _guard(
            _worm.worm_flash_health_summary(
                self._ctx,
                byref(val_errors),
                byref(val_spare),
                byref(val_erase),
                byref(val_retention),
            )
        )
        needs_replacement = bool(
            _worm.worm_flash_health_needs_replacement(val_errors, val_spare, val_erase)
        )
        return {
            "uncorrectableEccErrors": val_errors.value,
            "percentageRemainingSpareBlocks": val_spare.value,
            "percentageRemainingEraseCounts": val_erase.value,
            "percentageRemainingTenYearsDataRetention": val_retention.value,
            "needsReplacement": needs_replacement,
        }

    @log_execution
    def run_self_test(self, client_id):
        _guard(_worm.worm_tse_runSelfTest(self._ctx, client_id))

    @log_execution
    def factory_reset(self):
        # Works only on development TSEs (pre-2020).
        _guard(_worm.worm_tse_factoryReset(self._ctx))

    @log_execution
    def setup(
        self,
        client_id: str,
        admin_pin: str,
        admin_puk: str,
        time_admin_pin: str,
        credential_seed="SwissbitSwissbit",
    ):
        assert len(admin_pin) == 5
        assert len(admin_puk) == 6
        assert len(time_admin_pin) == 5
        _guard(
            _worm.worm_tse_setup(
                self._ctx,
                _c_ubyte(credential_seed.encode()),
                len(credential_seed),
                _c_ubyte(admin_puk.encode()),
                len(admin_puk),
                _c_ubyte(admin_pin.encode()),
                len(admin_pin),
                _c_ubyte(time_admin_pin.encode()),
                len(time_admin_pin),
                client_id,
            )
        )

    @log_execution
    def ctss_enable(self):
        _guard(_worm.worm_tse_ctss_enable(self._ctx))

    @log_execution
    def ctss_disable(self):
        _guard(_worm.worm_tse_ctss_disable(self._ctx))

    @log_execution
    def initialize(self):
        _guard(_worm.worm_tse_initialize(self._ctx))

    @log_execution
    def decommission(self):
        _guard(_worm.worm_tse_decommission(self._ctx))

    @log_execution
    def update_time(self, time=None):
        if not time:
            time = datetime.datetime.now(datetime.timezone.utc)
        if not isinstance(time, int):
            time = int(time.astimezone(datetime.timezone.utc).timestamp())
        _guard(_worm.worm_tse_updateTime(self._ctx, time))

    @log_execution
    def bundled_firmware_update_available(self) -> bool:
        val = _worm.WormTseFirmwareUpdate()
        _guard(_worm.worm_tse_firmwareUpdate_isBundledAvailable(self._ctx, byref(val)))
        return val.value != _worm.WORM_FW_NONE

    @log_execution
    def bundled_firmware_update_apply(self):
        _guard(_worm.worm_tse_firmwareUpdate_applyBundled(self._ctx))

    @log_execution
    def enable_export_if_csp_test_fails(self):
        _guard(_worm.worm_tse_enableExportIfCspTestFails(self._ctx))

    @log_execution
    def disable_export_if_csp_test_fails(self):
        _guard(_worm.worm_tse_disableExportIfCspTestFails(self._ctx))

    @log_execution
    def register_client(self, client_id: str):
        _guard(_worm.worm_tse_registerClient(self._ctx, client_id))

    @log_execution
    def deregister_client(self, client_id: str):
        _guard(_worm.worm_tse_deregisterClient(self._ctx, client_id))

    @log_execution
    def list_registered_clients(self) -> List[str]:
        result = []
        skip = 0
        while True:
            clients = _worm.WormRegisteredClients()
            _guard(
                _worm.worm_tse_listRegisteredClients(self._ctx, skip, byref(clients))
            )
            result += [
                c.raw.decode().rstrip("\x00")
                for c in clients.clientIds[: clients.amount]
            ]
            if clients.amount == 16:
                skip += 16
            else:
                break
        return result

    def _login(self, user: _worm.WormUserId, pin: str):
        val_remaining_retries = c_int()
        ret = _worm.worm_user_login(
            self._ctx,
            user,
            _c_ubyte(pin.encode()),
            len(pin.encode()),
            byref(val_remaining_retries),
        )
        if ret:
            logger.warning("Remaining retries: %d", val_remaining_retries.value)
        _guard(ret)

    @log_execution
    def login_as_admin(self, pin: str):
        self._login(_worm.WORM_USER_ADMIN, pin)

    @log_execution
    def login_as_time_admin(self, pin: str):
        self._login(_worm.WORM_USER_TIME_ADMIN, pin)

    @log_execution
    def logout_as_admin(self):
        _guard(_worm.worm_user_logout(self._ctx, _worm.WORM_USER_ADMIN))

    @log_execution
    def logout_as_time_admin(self):
        _guard(_worm.worm_user_logout(self._ctx, _worm.WORM_USER_TIME_ADMIN))

    def _unblock(self, user: _worm.WormUserId, puk: str, new_pin: str):
        val_remaining_retries = c_int()
        assert len(new_pin) == 5
        assert len(puk) == 6
        ret = _worm.worm_user_unblock(
            self._ctx,
            user,
            _c_ubyte(puk.encode()),
            len(puk),
            _c_ubyte(new_pin.encode()),
            len(new_pin),
            byref(val_remaining_retries),
        )
        if ret:
            logger.warning("Remaining retries: %d", val_remaining_retries.value)
        _guard(ret)

    @log_execution
    def unblock_admin(self, puk: str, new_pin: str):
        self._unblock(_worm.WORM_USER_ADMIN, puk, new_pin)

    @log_execution
    def unblock_time_admin(self, puk: str, new_pin: str):
        self._unblock(_worm.WORM_USER_TIME_ADMIN, puk, new_pin)

    @log_execution
    def change_puk(self, puk: str, new_puk: str):
        val_remaining_retries = c_int()
        assert len(puk) == 6
        assert len(new_puk) == 6
        ret = _worm.worm_user_change_puk(
            self._ctx,
            _c_ubyte(puk.encode()),
            len(puk),
            _c_ubyte(new_puk.encode()),
            len(new_puk),
            byref(val_remaining_retries),
        )
        if ret:
            logger.warning("Remaining retries: %d", val_remaining_retries.value)
        _guard(ret)

    def _change_pin(self, user: _worm.WormUserId, pin: str, new_pin: str):
        val_remaining_retries = c_int()
        assert len(new_pin) == 5
        assert len(pin) == 5
        ret = _worm.worm_user_change_pin(
            self._ctx,
            user,
            _c_ubyte(pin.encode()),
            len(pin),
            _c_ubyte(new_pin.encode()),
            len(new_pin),
            byref(val_remaining_retries),
        )
        if ret:
            logger.warning("Remaining retries: %d", val_remaining_retries.value)
        _guard(ret)

    @log_execution
    def change_admin_pin(self, pin: str, new_pin: str):
        self._change_pin(_worm.WORM_USER_ADMIN, pin, new_pin)

    @log_execution
    def change_time_admin_pin(self, pin: str, new_pin: str):
        self._change_pin(_worm.WORM_USER_TIME_ADMIN, pin, new_pin)

    @log_execution
    def derive_initial_credentials(self, credential_seed="SwissbitSwissbit") -> dict:
        _worm.worm_user_deriveInitialCredentials.argtypes = [
            # This is easier to handle than what ctypesgen generates
            POINTER(_worm.WormContext),
            POINTER(c_ubyte),
            c_int,
            c_char_p,
            c_int,
            c_char_p,
            c_int,
            c_char_p,
            c_int,
        ]
        val_res_puk = c_char_p(b"******")
        val_res_pin = c_char_p(b"*****")
        val_res_tapin = c_char_p(b"*****")

        _guard(
            _worm.worm_user_deriveInitialCredentials(
                self._ctx,
                _c_ubyte(credential_seed.encode()),
                len(credential_seed),
                val_res_puk,
                6,
                val_res_pin,
                5,
                val_res_tapin,
                5,
            )
        )
        return {
            "adminPuk": val_res_puk.value.decode(),
            "adminPin": val_res_pin.value.decode(),
            "timeAdminPin": val_res_tapin.value.decode(),
        }

    def _transaction_response_to_dict(self, resp):
        val_res = POINTER(c_ubyte)()
        val_len = _worm.worm_uint()
        _worm.worm_transaction_response_serialNumber(
            resp, byref(val_res), byref(val_len)
        )
        serial = bytes([val_res[i] for i in range(val_len.value)])

        val_res = POINTER(c_ubyte)()
        val_len = _worm.worm_uint()
        _worm.worm_transaction_response_signature(resp, byref(val_res), byref(val_len))
        signature = bytes([val_res[i] for i in range(val_len.value)])

        return {
            "logTime": _worm.worm_transaction_response_logTime(resp),
            "serialNumberHex": serial.hex(),
            "serialNumberBytes": serial,
            "signatureCounter": _worm.worm_transaction_response_signatureCounter(resp),
            "transactionNumber": _worm.worm_transaction_response_transactionNumber(
                resp
            ),
            "signatureBytes": signature,
            "signatureBase64": base64.b64encode(signature).decode(),
        }

    @log_execution
    def transaction_start(self, client_id: str, process_data: str, process_type: str):
        resp = _worm.worm_transaction_response_new(self._ctx)
        _guard(
            _worm.worm_transaction_start(
                self._ctx,
                client_id,
                _c_ubyte(process_data.encode()),
                len(process_data.encode()),
                process_type,
                resp,
            )
        )
        data = self._transaction_response_to_dict(resp)
        _worm.worm_transaction_response_free(resp)
        return data

    @log_execution
    def transaction_update(
        self, client_id: str, transaction_id: int, process_data: str, process_type: str
    ):
        resp = _worm.worm_transaction_response_new(self._ctx)
        _guard(
            _worm.worm_transaction_update(
                self._ctx,
                client_id,
                transaction_id,
                _c_ubyte(process_data.encode()),
                len(process_data.encode()),
                process_type,
                resp,
            )
        )
        data = self._transaction_response_to_dict(resp)
        _worm.worm_transaction_response_free(resp)
        return data

    @log_execution
    def transaction_finish(
        self, client_id: str, transaction_id: int, process_data: str, process_type: str
    ):
        resp = _worm.worm_transaction_response_new(self._ctx)
        _guard(
            _worm.worm_transaction_finish(
                self._ctx,
                client_id,
                transaction_id,
                _c_ubyte(process_data.encode()),
                len(process_data.encode()),
                process_type,
                resp,
            )
        )
        data = self._transaction_response_to_dict(resp)
        _worm.worm_transaction_response_free(resp)
        return data

    @log_execution
    def last_transaction(self, client_id: str = None):
        resp = _worm.worm_transaction_response_new(self._ctx)
        _guard(
            _worm.worm_transaction_lastResponse(
                self._ctx,
                client_id,
                resp,
            )
        )
        data = self._transaction_response_to_dict(resp)
        _worm.worm_transaction_response_free(resp)
        return data

    @log_execution
    def list_started_transactions(self, client_id: str = None):
        result = []
        skip = 0
        while True:
            val_res = cast((_worm.worm_uint * 62)(), POINTER(_worm.worm_uint))
            val_len = c_int()
            _guard(
                _worm.worm_transaction_listStartedTransactions(
                    self._ctx, client_id, skip, val_res, 62, byref(val_len)
                )
            )
            result += [val_res[i] for i in range(val_len.value)]
            if val_len.value == 62:
                skip += 62
            else:
                break
        return result

    def _entry_to_dict(self, entry: _worm.WormEntry) -> dict:
        types = {
            _worm.WORM_ENTRY_TYPE_TRANSACTION: "WORM_ENTRY_TYPE_TRANSACTION",
            _worm.WORM_ENTRY_TYPE_SYSTEM_LOG_MESSAGE: "WORM_ENTRY_TYPE_SYSTEM_LOG_MESSAGE",
            _worm.WORM_ENTRY_TYPE_SE_AUDIT_LOG_MESSAGE: "WORM_ENTRY_TYPE_SE_AUDIT_LOG_MESSAGE",
        }
        val_len = _worm.worm_entry_logMessageLength(entry)
        val_res = cast((c_ubyte * val_len)(), POINTER(c_ubyte))
        _guard(_worm.worm_entry_readLogMessage(entry, val_res, val_len))

        val_len_pd = _worm.worm_entry_processDataLength(entry)
        val_res_pd = cast((c_ubyte * val_len)(), POINTER(c_ubyte))
        _guard(_worm.worm_entry_readProcessData(entry, 0, val_res_pd, val_len_pd))

        return {
            "isValid": bool(_worm.worm_entry_isValid(entry)),
            "id": _worm.worm_entry_id(entry),
            "type": types[_worm.worm_entry_type(entry)],
            "message": bytes([val_res[i] for i in range(val_len)]),
            "processData": bytes([val_res_pd[i] for i in range(val_len_pd)]),
        }

    @log_execution
    def first_entry(self) -> dict:
        entry = _worm.worm_entry_new(self._ctx)
        _guard(_worm.worm_entry_iterate_first(entry))
        d = self._entry_to_dict(entry)
        _worm.worm_entry_free(entry)
        return d

    @log_execution
    def last_entry(self) -> dict:
        entry = _worm.worm_entry_new(self._ctx)
        _guard(_worm.worm_entry_iterate_last(entry))
        d = self._entry_to_dict(entry)
        _worm.worm_entry_free(entry)
        return d

    @log_execution
    def entry_by_id(self, entry_id: int) -> dict:
        entry = _worm.worm_entry_new(self._ctx)
        _guard(_worm.worm_entry_iterate_id(entry, entry_id))
        d = self._entry_to_dict(entry)
        _worm.worm_entry_free(entry)
        return d

    @log_execution
    def iterate_entries(self, start_at: int = 0):
        entry = _worm.worm_entry_new(self._ctx)
        if start_at:
            _guard(_worm.worm_entry_iterate_id(entry, start_at))
        else:
            _guard(_worm.worm_entry_iterate_first(entry))
        try:
            if _worm.worm_entry_isValid(entry):
                yield self._entry_to_dict(entry)

                while True:
                    _guard(_worm.worm_entry_iterate_next(entry))
                    if not _worm.worm_entry_isValid(entry):
                        break
                    yield self._entry_to_dict(entry)
        finally:
            _worm.worm_entry_free(entry)

    @log_execution
    def get_log_message_certificate(self) -> str:
        val_len = c_uint(
            1024 * 16
        )  # 16 kb buffer should be enough. it if is not, the function will fail.
        val_res = cast((c_ubyte * val_len.value)(), POINTER(c_ubyte))
        _guard(_worm.worm_getLogMessageCertificate(self._ctx, val_res, byref(val_len)))
        return bytes([val_res[i] for i in range(val_len.value)]).decode()

    @log_execution
    def export_tar(
        self,
        target: BinaryIO,
        start_date: datetime.datetime = None,
        end_date: datetime.datetime = None,
        client_id: str = None,
        start_transaction: int = None,
        end_transaction: int = None,
    ):
        def py_write(chunk, chunk_length, callback_data):
            data = bytes([chunk[i] for i in range(chunk_length)])
            target.write(data)
            return 0

        write = _worm.WormExportTarCallback(py_write)

        if start_date is not None or end_date is not None:
            if start_transaction or end_transaction:
                raise ValueError("You can either filter by date or transaction")
            if isinstance(start_date, datetime.datetime):
                start_date = int(
                    start_date.astimezone(datetime.timezone.utc).timestamp()
                )
            elif start_date is None:
                start_date = None
            if isinstance(end_date, datetime.datetime):
                end_date = int(end_date.astimezone(datetime.timezone.utc).timestamp())
            elif end_date is None:
                end_date = 0xFFFFFFFFFFFFFFFF

            _guard(
                _worm.worm_export_tar_filtered_time(
                    self._ctx, start_date, end_date, client_id, write, c_void_p()
                )
            )
        elif start_transaction is not None or end_transaction is not None:
            if start_date or end_date:
                raise ValueError("You can either filter by date or transaction")

            if start_transaction is None:
                start_transaction = 0
            if end_transaction is None:
                end_transaction = 0xFFFFFFFFFFFFFFFF

            _guard(
                _worm.worm_export_tar_filtered_transaction(
                    self._ctx,
                    start_transaction,
                    end_transaction,
                    client_id,
                    write,
                    c_void_p(),
                )
            )
        else:
            if client_id:
                _guard(
                    _worm.worm_export_tar_filtered_time(
                        self._ctx, 0, 0xFFFFFFFFFFFFFFFF, client_id, write, c_void_p()
                    )
                )
            else:
                _guard(_worm.worm_export_tar(self._ctx, write, c_void_p()))

    @log_execution
    def delete_stored_data(self):
        _guard(_worm.worm_export_deleteStoredData(self._ctx))


class LocalWormContext(BaseWormContext):
    @log_execution
    def __init__(self, mount_point: str):
        super().__init__()
        self._ctx = POINTER(_worm.WormContext)()
        _guard(_worm.worm_init(byref(self._ctx), mount_point.encode()))


class LANWormContext(BaseWormContext):
    @log_execution
    def __init__(self, url: str, api_token: str):
        super().__init__()
        self._ctx = POINTER(_worm.WormContext)()
        _guard(_worm.worm_init_lan(byref(self._ctx), url, api_token))

    def select_tse(self, serial_number: str):
        _guard(
            _worm.worm_lantse_select(
                self._ctx,
                c_ubyte(serial_number.encode()),
                len(serial_number.encode()),
            )
        )

    @contextlib.contextmanager
    def lock_tse(self):
        _guard(_worm.worm_lantse_lock(self._ctx))
        try:
            yield
        finally:
            _guard(_worm.worm_lantse_unlock(self._ctx))

    def list_connected_tses(self) -> List[str]:
        result = []
        skip = 0
        while True:
            tses = _worm.WormSerialNumberList()
            _guard(_worm.worm_lantse_listConnectedTses(self._ctx, skip, byref(tses)))
            result += [
                c.raw.decode().rstrip("\x00") for c in tses.serialNumber[: tses.amount]
            ]
            if tses.amount == 16:
                skip += 16
            else:
                break
        return result



import datetime
import os
from contextlib import contextmanager

import click
import dateutil.parser
import uvicorn
from click import UsageError
from rich.console import Console
from rich.table import Table

from sbtse import worm
from sbtse.errors import (
    WormError,
    WormErrorClientNotRegistered,
    WormErrorWrongStateNeedsSelfTest,
)


class PinParamType(click.ParamType):
    name = "pin"

    def convert(self, value, param, ctx):
        if len(value) != 5:
            self.fail(f"{value!r} is not of length 5", param, ctx)
        if not value.isdigit():
            self.fail(f"{value!r} is not only digits", param, ctx)
        return value


class PukParamType(click.ParamType):
    name = "puk"

    def convert(self, value, param, ctx):
        if len(value) != 6:
            self.fail(f"{value!r} is not of length 6", param, ctx)
        if not value.isdigit():
            self.fail(f"{value!r} is not only digits", param, ctx)
        return value


class DateTimeParamType(click.ParamType):
    name = "datetime"

    def convert(self, value, param, ctx):
        return dateutil.parser.parse(value)


T_PIN = PinParamType()
T_PUK = PukParamType()
T_DATETIME = DateTimeParamType()


@click.group()
@click.option(
    "--path",
    type=click.Path(
        exists=True, dir_okay=True, file_okay=False, readable=True, writable=True
    ),
    help="TSE mount point",
)
@click.option("--url", type=str, help="LAN TSE URL")
@click.option("--api-key", type=str, help="LAN TSE API Key")
@click.option("--tse", type=str, help="LAN TSE serial number")
@click.pass_context
def main(ctx, path, url, api_key, tse):
    ctx.ensure_object(dict)
    if not path and not (url and api_key and tse) and ctx.invoked_subcommand != "mock":
        raise UsageError("Please set either --path or --url/--api-key/--tse.")
    ctx.obj["path"] = path
    ctx.obj["url"] = url
    ctx.obj["api_key"] = api_key
    ctx.obj["tse"] = tse


@contextmanager
def _tse_context(ctx, admin_pin=None, time_admin_pin=None, self_test_client=None):
    if ctx.obj["path"]:
        with worm.LocalWormContext(ctx.obj["path"]) as w:
            info = w.info()
            if not info["hasPassedSelfTest"] and self_test_client:
                try:
                    w.run_self_test(self_test_client)
                except WormErrorClientNotRegistered:
                    pass
            if admin_pin:
                w.login_as_admin(admin_pin)
            if time_admin_pin:
                w.login_as_time_admin(time_admin_pin)
                if info["isDevelopmentFirmware"] and info[
                    "certificateExpirationDate"
                ] < datetime.datetime.now(datetime.timezone.utc):
                    # Development TSEs expired 2020-01, so let's backdate so we can still use them for testing
                    w.update_time(
                        datetime.datetime.now(datetime.timezone.utc).replace(year=2019)
                    )
                else:
                    w.update_time()
            yield w
    else:
        with worm.LANWormContext(ctx.obj["url"], ctx.obj["api_key"]) as w:
            info = w.info()
            w.select_tse(ctx.obj["tse"])
            if not info["hasPassedSelfTest"] and self_test_client:
                try:
                    w.run_self_test(self_test_client)
                except WormErrorClientNotRegistered:
                    pass
            if admin_pin:
                w.login_as_admin(admin_pin)
            if time_admin_pin:
                w.login_as_time_admin(time_admin_pin)
                if info["isDevelopmentFirmware"] and info[
                    "certificateExpirationDate"
                ] < datetime.datetime.now(datetime.timezone.utc):
                    # Development TSEs expired 2020-01, so let's backdate so we can still use them for testing
                    w.update_time(
                        datetime.datetime.now(datetime.timezone.utc).replace(year=2019)
                    )
                else:
                    w.update_time()
            yield w


@main.command(help="Show info and flash health status")
@click.pass_context
def info(ctx):
    with _tse_context(ctx) as w:
        info = w.info()

        table = Table(title="Information")
        table.add_column("Key", justify="left", no_wrap=True)
        table.add_column("Value", justify="left")
        for k, v in info.items():
            table.add_row(k, str(v))

        console = Console()
        console.print(table)

        try:
            info = w.flash_health()
        except WormErrorWrongStateNeedsSelfTest:
            click.echo(
                click.style("Flash health can only be shown after selftest.", fg="red")
            )
            return

        table = Table(title="Flash health")
        table.add_column("Key", justify="left", no_wrap=True)
        table.add_column("Value", justify="left")
        for k, v in info.items():
            table.add_row(k, str(v))

        console = Console()
        console.print(table)


@main.command(help="Run self-test")
@click.option("--client-id", "-c", prompt=True, type=str, help="Client ID")
@click.pass_context
def selftest(ctx, client_id):
    with _tse_context(ctx) as w:
        try:
            w.run_self_test(client_id)
        except WormError as e:
            click.echo(click.style(f"Failed with error: {type(e).__name__}", fg="red"))
        else:
            click.echo(click.style("OK", fg="green"))


@main.command(help="Factory reset (development TSE only)")
@click.pass_context
def factory_reset(ctx):
    with _tse_context(ctx) as w:
        w.factory_reset()


@main.command(help="Run setup procedure for a fresh TSE")
@click.option("--client-id", "-c", prompt=True, type=str, help="Client ID")
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.option("--admin-puk", prompt=True, type=T_PUK, help="Admin PUK")
@click.option("--time-admin-pin", prompt=True, type=T_PIN, help="Time Admin PIN")
@click.pass_context
def setup(ctx, client_id, admin_pin, admin_puk, time_admin_pin):
    with _tse_context(ctx) as w:
        try:
            w.run_self_test(client_id)
        except WormErrorClientNotRegistered:
            pass
        else:
            raise UsageError("Client already registered.")
        w.setup(client_id, admin_pin, admin_puk, time_admin_pin)


@main.group("config", help="Manipulate TSE configuration")
@click.pass_context
def config(ctx):
    pass


@config.command(help="Enable CTSS")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selftest)"
)
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def ctss_enable(ctx, client_id, admin_pin):
    with _tse_context(ctx, self_test_client=client_id, admin_pin=admin_pin) as w:
        w.ctss_enable()


@config.command(help="Disable CTSS")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selfest)"
)
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def ctss_disable(ctx, client_id, admin_pin):
    with _tse_context(ctx, self_test_client=client_id, admin_pin=admin_pin) as w:
        w.ctss_disable()


@config.command(help="Final step for TSE initialization")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selftest)"
)
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def initialize(ctx, client_id, admin_pin):
    with _tse_context(ctx, self_test_client=client_id, admin_pin=admin_pin) as w:
        w.initialize()


@config.command(help="Register client")
@click.option("--client-id", "-c", prompt=True, type=str, help="Client ID")
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def register_client(ctx, client_id, admin_pin):
    with _tse_context(ctx, self_test_client=client_id, admin_pin=admin_pin) as w:
        w.register_client(client_id)


@config.command(help="Deregister client")
@click.option("--client-id", "-c", prompt=True, type=str, help="Client ID")
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def deregister_client(ctx, client_id, admin_pin):
    with _tse_context(ctx, self_test_client=client_id, admin_pin=admin_pin) as w:
        w.deregister_client(client_id)


@config.command(help="List of registered clients")
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def list_clients(ctx, admin_pin):
    with _tse_context(ctx, admin_pin=admin_pin) as w:
        for c in w.list_registered_clients():
            print(c)


@config.command(help="Enable export if CSP test fails")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selftest)"
)
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def enable_export_if_csp_test_fails(ctx, client_id, admin_pin):
    with _tse_context(ctx, self_test_client=client_id, admin_pin=admin_pin) as w:
        w.enable_export_if_csp_test_fails()


@config.command(help="Disable export if CSP test fails")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selftest)"
)
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def disable_export_if_csp_test_fails(ctx, client_id, admin_pin):
    with _tse_context(ctx, self_test_client=client_id, admin_pin=admin_pin) as w:
        w.disable_export_if_csp_test_fails()


@config.command(help="Decommission (irrevocably disable) TSE")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selftest)"
)
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def decommission(ctx, client_id, admin_pin):
    with _tse_context(ctx, self_test_client=client_id, admin_pin=admin_pin) as w:
        w.update_time()
        w.decommission()


@main.command(help="Update firmware to version bundled with SDK.")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selftest)"
)
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.pass_context
def firmware_update(ctx, client_id, admin_pin):
    with _tse_context(ctx, self_test_client=client_id, admin_pin=admin_pin) as w:
        if w.bundled_firmware_update_available():
            click.echo("Applying firmware update...")
            w.bundled_firmware_update_apply()
            click.echo(click.style("OK", fg="green"))
        else:
            click.echo(click.style("No update available.", fg="red"))


@main.group("pin", help="Manage Admin PIN")
@click.pass_context
def pin(ctx):
    pass


@pin.command(help="Unblock Admin PIN")
@click.option("--admin-puk", prompt=True, type=T_PUK, help="Admin PUK")
@click.option("--new-pin", prompt=True, type=T_PIN, help="New PIN")
@click.pass_context
def unblock(ctx, admin_puk, new_pin):
    with _tse_context(ctx, self_test_client="UNKNOWN") as w:
        w.unblock_admin(admin_puk, new_pin)


@pin.command(help="Change Admin PIN")
@click.option("--old-pin", prompt=True, type=T_PIN, help="Old PIN")
@click.option("--new-pin", prompt=True, type=T_PIN, help="New PIN")
@click.pass_context
def change(ctx, old_pin, new_pin):
    with _tse_context(ctx, self_test_client="UNKNOWN") as w:
        w.change_admin_pin(old_pin, new_pin)


@main.group("time-admin-pin", help="Manage Time Admin PIN")
@click.pass_context
def time_admin_pin(ctx):
    pass


@time_admin_pin.command(help="Unblock Time Admin PIN")
@click.option("--admin-puk", prompt=True, type=T_PUK, help="Admin PUK")
@click.option("--new-pin", prompt=True, type=T_PIN, help="New PIN")
@click.pass_context
def unblock(ctx, admin_puk, new_pin):
    with _tse_context(ctx, self_test_client="UNKNOWN") as w:
        w.unblock_time_admin(admin_puk, new_pin)


@time_admin_pin.command(help="Change Time Admin PIN")
@click.option("--old-pin", prompt=True, type=T_PIN, help="Old PIN")
@click.option("--new-pin", prompt=True, type=T_PIN, help="New PIN")
@click.pass_context
def change(ctx, old_pin, new_pin):
    with _tse_context(ctx, self_test_client="UNKNOWN") as w:
        w.change_time_admin_pin(old_pin, new_pin)


@main.group("puk", help="Manage PUK")
@click.pass_context
def puk(ctx):
    pass


@puk.command(help="Change PUK")
@click.option("--old-puk", prompt=True, type=T_PUK, help="Old PUK")
@click.option("--new-puk", prompt=True, type=T_PUK, help="New PUK")
@click.pass_context
def change(ctx, old_puk, new_puk):
    with _tse_context(ctx, self_test_client="UNKNOWN") as w:
        w.change_puk(old_puk, new_puk)


@puk.command(help="Show initial credentials")
@click.pass_context
def initial(ctx):
    with _tse_context(ctx, self_test_client="UNKNOWN") as w:
        for k, v in w.derive_initial_credentials().items():
            print(f"{k}: {v}")


@main.group("transaction", help="Create and query transactions")
@click.pass_context
def transaction(ctx):
    pass


def _print_tx(tx):
    table = Table(title="Transaction")
    table.add_column("Key", justify="left", no_wrap=True)
    table.add_column("Value", justify="left")
    for k, v in tx.items():
        table.add_row(k, str(v))

    console = Console()
    console.print(table)


@transaction.command(help="Start transaction")
@click.option("--time-admin-pin", prompt=True, type=T_PIN, help="Time Admin PIN")
@click.argument("client_id")
@click.argument("process_type", required=False)
@click.argument("process_data", required=False)
@click.pass_context
def start(ctx, client_id, process_type, process_data, time_admin_pin):
    with _tse_context(
        ctx, self_test_client=client_id, time_admin_pin=time_admin_pin
    ) as w:
        tx = w.transaction_start(client_id, process_data or "", process_type or "")
        _print_tx(tx)


@transaction.command(help="Update transaction")
@click.option("--time-admin-pin", prompt=True, type=T_PIN, help="Time Admin PIN")
@click.argument("client_id")
@click.argument("transaction_id", type=int)
@click.argument("process_type")
@click.argument("process_data")
@click.pass_context
def update(ctx, client_id, transaction_id, process_type, process_data, time_admin_pin):
    with _tse_context(
        ctx, self_test_client=client_id, time_admin_pin=time_admin_pin
    ) as w:
        tx = w.transaction_update(client_id, transaction_id, process_data, process_type)
        _print_tx(tx)


@transaction.command(help="Finish transaction")
@click.option("--time-admin-pin", prompt=True, type=T_PIN, help="Time Admin PIN")
@click.argument("client_id")
@click.argument("transaction_id", type=int)
@click.argument("process_type")
@click.argument("process_data")
@click.pass_context
def finish(ctx, client_id, transaction_id, process_type, process_data, time_admin_pin):
    with _tse_context(
        ctx, self_test_client=client_id, time_admin_pin=time_admin_pin
    ) as w:
        tx = w.transaction_finish(client_id, transaction_id, process_data, process_type)
        _print_tx(tx)


@transaction.command(help="Show details of last transaction")
@click.argument("client_id", required=False)
@click.pass_context
def last(ctx, client_id):
    with _tse_context(ctx, self_test_client=client_id) as w:
        _print_tx(w.last_transaction(client_id))


@transaction.command(help="List started transaction IDs")
@click.argument("client_id", required=False)
@click.pass_context
def list_started(ctx, client_id):
    with _tse_context(ctx, self_test_client=client_id) as w:
        for tx in w.list_started_transactions(client_id):
            print(tx)


@main.group("entries", help="Query log entries")
@click.pass_context
def entries(ctx):
    pass


@entries.command(help="Retrieve entry by ID")
@click.argument("entry_id", type=int)
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selftest)"
)
@click.pass_context
def get(ctx, admin_pin, client_id, entry_id):
    with _tse_context(ctx, admin_pin=admin_pin, self_test_client=client_id) as w:
        entry = w.entry_by_id(entry_id)
        table = Table(title="Flash health")
        table.add_column("Key", justify="left", no_wrap=True)
        table.add_column("Value", justify="left")
        for k, v in entry.items():
            table.add_row(k, str(v))

        console = Console()
        console.print(table)


@entries.command(help="List entries")
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selftest)"
)
@click.pass_context
def list(ctx, admin_pin, client_id):
    with _tse_context(ctx, admin_pin=admin_pin, self_test_client=client_id) as w:
        table = Table(title="Entries")
        table.add_column("ID", justify="left", no_wrap=True)
        table.add_column("Valid", justify="left")
        table.add_column("Type", justify="left")
        table.add_column("Message", justify="left")
        table.add_column("Process Data", justify="left")
        for entry in w.iterate_entries():
            table.add_row(
                str(entry["id"]),
                str(entry["isValid"]),
                entry["type"],
                str(entry["message"]),
                str(entry["processData"]),
            )

        console = Console()
        console.print(table)


@main.command(help="Delete stored data. Requires a full export first.")
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.option(
    "--client-id", "-c", prompt=True, type=str, help="Client ID (for selftest)"
)
@click.pass_context
def delete(ctx, admin_pin, client_id):
    with _tse_context(ctx, admin_pin=admin_pin, self_test_client=client_id) as w:
        w.delete_stored_data()


@main.command(help="Export stored data.")
@click.argument("outfile", type=click.File("wb"))
@click.option("--admin-pin", prompt=True, type=T_PIN, help="Admin PIN")
@click.option(
    "--client-id",
    "-c",
    prompt=True,
    type=str,
    help="Client ID (for selftest, if necessary)",
)
@click.option("--filter-client-id", type=str, help="Client ID (for filtering)")
@click.option("--start-date", type=T_DATETIME, help="Start date (inclusively)")
@click.option("--end-date", type=T_DATETIME, help="End date (inclusively)")
@click.option("--start-transaction", type=int, help="Start transaction (inclusive)")
@click.option("--end-transaction", type=int, help="End transaction (inclusive)")
@click.pass_context
def export(
    ctx,
    outfile,
    admin_pin,
    client_id,
    filter_client_id,
    start_date,
    end_date,
    start_transaction,
    end_transaction,
):
    with _tse_context(ctx, admin_pin=admin_pin, self_test_client=client_id) as w:
        w.export_tar(
            outfile,
            start_date=start_date,
            end_date=end_date,
            client_id=filter_client_id,
            start_transaction=start_transaction,
            end_transaction=end_transaction,
        )


@main.command(help="Run local API server")
@click.option("--reload", is_flag=True, help="Auto-reload code (development only)")
@click.option("--debug", is_flag=True, help="Debug log")
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Host to listen to (default: 127.0.0.1)",
)
@click.option(
    "--port", type=int, default=9873, help="Host to listen to (default: 9873)"
)
@click.option("--time-admin-pin", prompt=True, type=T_PIN, help="Time Admin PIN")
@click.pass_context
def serve(ctx, host, port, reload, time_admin_pin, debug):
    _serve(ctx, host, port, reload, time_admin_pin, debug, "sbtse.api:app")


@main.command(help="Run mock API server (development only)")
@click.option("--reload", is_flag=True, help="Auto-reload code (development only)")
@click.option("--debug", is_flag=True, help="Debug log")
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Host to listen to (default: 127.0.0.1)",
)
@click.option(
    "--port", type=int, default=9873, help="Host to listen to (default: 9873)"
)
@click.pass_context
def mock(ctx, host, port, reload, debug):
    _serve(ctx, host, port, reload, "12345", debug, "sbtse.mockapi:app")


def _serve(ctx, host, port, reload, time_admin_pin, debug, app):
    os.environ.update(
        {
            "SBTSE_PATH": ctx.obj["path"] or "",
            "SBTSE_URL": ctx.obj["url"] or "",
            "SBTSE_API_KEY": ctx.obj["api_key"] or "",
            "SBTSE_TSE": ctx.obj["tse"] or "",
            "SBTSE_TAPIN": time_admin_pin,
        }
    )
    log_config = uvicorn.config.LOGGING_CONFIG
    log_config["formatters"]["access"][
        "fmt"
    ] = "%(asctime)s - %(levelname)s - %(name)s - thread=%(thread)s - %(message)s"
    log_config["formatters"]["default"][
        "fmt"
    ] = "%(asctime)s - %(levelname)s - %(name)s - thread=%(thread)s - %(message)s"
    log_config["loggers"][""] = {
        "handlers": ["default"],
        "level": "DEBUG" if debug else "INFO",
        "propagate": False,
    }
    uvicorn.run(
        app,
        host=host,
        port=port,
        workers=1,  # No concurrent access to TSE allowed!
        reload=reload,
        log_config=log_config,
    )


if __name__ == "__main__":
    main(obj={})



class WormError(Exception):
    pass


class WormErrorInvalidParameter(WormError):
    pass


class WormErrorNoWormCard(WormError):
    pass


class WormErrorIO(WormError):
    pass


class WormErrorTimeout(WormError):
    pass


class WormErrorOutOfMem(WormError):
    pass


class WormErrorInvalidResponse(WormError):
    pass


class WormErrorStoreFullInternal(WormError):
    pass


class WormErrorResponseMissing(WormError):
    pass


class WormErrorExportNotInitialized(WormError):
    pass


class WormErrorExportFailed(WormError):
    pass


class WormErrorIncrementalExportInvalidState(WormError):
    pass


class WormErrorIncrementalExportNoData(WormError):
    pass


class WormErrorPowerCycleDetected(WormError):
    pass


class WormErrorFirmwareUpdateNotApplied(WormError):
    pass


class WormErrorThreadStartFailed(WormError):
    pass


class WormErrorNetwork(WormError):
    pass


class WormErrorCmdNotSupported(WormError):
    pass


class WormErrorLanInvalidApiToken(WormError):
    pass


class WormErrorNetworkTimeout(WormError):
    pass


class WormErrorConnectionFailed(WormError):
    pass


class WormErrorLanUnbalancedLocks(WormError):
    pass


class WormErrorLanInvalidServerResponse(WormError):
    pass


class WormErrorInvalidState(WormError):
    pass


class WormErrorTseNotFound(WormError):
    pass


class WormErrorIncrementalExportLimitTooLow(WormError):
    pass


class WormErrorFwuNotAvailable(WormError):
    pass


class WormErrorFromCardFirst(WormError):
    pass


class WormErrorUnknown(WormError):
    pass


class WormErrorNoTimeSet(WormError):
    pass


class WormErrorNoTransactionInProgress(WormError):
    pass


class WormErrorInvalidCmdSyntax(WormError):
    pass


class WormErrorWrongLength(WormError):
    pass


class WormErrorNotEnoughDataWritten(WormError):
    pass


class WormErrorTseInvalidParameter(WormError):
    pass


class WormErrorTransactionNotStarted(WormError):
    pass


class WormErrorMaxParallelTransactions(WormError):
    pass


class WormErrorCertificateExpired(WormError):
    pass


class WormErrorNoLastTransaction(WormError):
    pass


class WormErrorCmdNotAllowed(WormError):
    pass


class WormErrorTransactionSignaturesExceeded(WormError):
    pass


class WormErrorNotAuthorized(WormError):
    pass


class WormErrorMaxRegisteredClientsReached(WormError):
    pass


class WormErrorClientNotRegistered(WormError):
    pass


class WormErrorExportUnacknowledgedData(WormError):
    pass


class WormErrorClientHasUnfinishedTransactions(WormError):
    pass


class WormErrorTseHasUnfinishedTransactions(WormError):
    pass


class WormErrorTseNoResponseToFetch(WormError):
    pass


class WormErrorNotAllowedExportInProgress(WormError):
    pass


class WormErrorStoreFull(WormError):
    pass


class WormErrorWrongStateNeedsPukChange(WormError):
    pass


class WormErrorWrongStateNeedsPinChange(WormError):
    pass


class WormErrorWrongStateNeedsActiveCtss(WormError):
    pass


class WormErrorWrongStateNeedsSelfTest(WormError):
    pass


class WormErrorWrongStateNeedsSelfTestPassed(WormError):
    pass


class WormErrorFwuIntegrityFailure(WormError):
    pass


class WormErrorFwuDecryptionFailure(WormError):
    pass


class WormErrorFwuWrongFormat(WormError):
    pass


class WormErrorFwuInternalError(WormError):
    pass


class WormErrorFwuDowngradeProhibited(WormError):
    pass


class WormErrorTseAlreadyInitialized(WormError):
    pass


class WormErrorTseDecommissioned(WormError):
    pass


class WormErrorTseNotInitialized(WormError):
    pass


class WormErrorAuthenticationFailed(WormError):
    pass


class WormErrorAuthenticationPinBlocked(WormError):
    pass


class WormErrorAuthenticationUserNotLoggedIn(WormError):
    pass


class WormErrorSelfTestFailedFW(WormError):
    pass


class WormErrorSelfTestFailedCSP(WormError):
    pass


class WormErrorSelfTestFailedRNG(WormError):
    pass


class WormErrorFwuBaseFWError(WormError):
    pass


class WormErrorFwuFwextError(WormError):
    pass


class WormErrorFwuCspError(WormError):
    pass


class WormErrorExportNoneInProgress(WormError):
    pass


class WormErrorExportRetry(WormError):
    pass


class WormErrorExportNoDataAvailable(WormError):
    pass


class WormErrorCmdNotFound(WormError):
    pass


class WormErrorSigError(WormError):
    pass


class WormErrorFromCardLast(WormError):
    pass


def _guard(rescode, allow=0):
    if rescode == allow:
        return
    if rescode == 1:
        raise WormErrorInvalidParameter()
    if rescode == 2:
        raise WormErrorNoWormCard()
    if rescode == 3:
        raise WormErrorIO()
    if rescode == 4:
        raise WormErrorTimeout()
    if rescode == 5:
        raise WormErrorOutOfMem()
    if rescode == 6:
        raise WormErrorInvalidResponse()
    if rescode == 7:
        raise WormErrorStoreFullInternal()
    if rescode == 8:
        raise WormErrorResponseMissing()
    if rescode == 9:
        raise WormErrorExportNotInitialized()
    if rescode == 10:
        raise WormErrorExportFailed()
    if rescode == 11:
        raise WormErrorIncrementalExportInvalidState()
    if rescode == 12:
        raise WormErrorIncrementalExportNoData()
    if rescode == 13:
        raise WormErrorPowerCycleDetected()
    if rescode == 14:
        raise WormErrorFirmwareUpdateNotApplied()
    if rescode == 15:
        raise WormErrorThreadStartFailed()
    if rescode == 16:
        raise WormErrorNetwork()
    if rescode == 17:
        raise WormErrorCmdNotSupported()
    if rescode == 18:
        raise WormErrorLanInvalidApiToken()
    if rescode == 19:
        raise WormErrorNetworkTimeout()
    if rescode == 20:
        raise WormErrorConnectionFailed()
    if rescode == 21:
        raise WormErrorLanUnbalancedLocks()
    if rescode == 22:
        raise WormErrorLanInvalidServerResponse()
    if rescode == 23:
        raise WormErrorInvalidState()
    if rescode == 24:
        raise WormErrorTseNotFound()
    if rescode == 25:
        raise WormErrorIncrementalExportLimitTooLow()
    if rescode == 26:
        raise WormErrorFwuNotAvailable()
    if rescode == 0x1000:
        raise WormErrorFromCardFirst()
    if rescode == 0x1001:
        raise WormErrorUnknown()
    if rescode == 0x1002:
        raise WormErrorNoTimeSet()
    if rescode == 0x1004:
        raise WormErrorNoTransactionInProgress()
    if rescode == 0x1005:
        raise WormErrorInvalidCmdSyntax()
    if rescode == 0x1006:
        raise WormErrorNotEnoughDataWritten()
    if rescode == 0x1007:
        raise WormErrorTseInvalidParameter()
    if rescode == 0x1008:
        raise WormErrorTransactionNotStarted()
    if rescode == 0x1009:
        raise WormErrorMaxParallelTransactions()
    if rescode == 0x100A:
        raise WormErrorCertificateExpired()
    if rescode == 0x100C:
        raise WormErrorNoLastTransaction()
    if rescode == 0x100D:
        raise WormErrorCmdNotAllowed()
    if rescode == 0x100E:
        raise WormErrorTransactionSignaturesExceeded()
    if rescode == 0x100F:
        raise WormErrorNotAuthorized()
    if rescode == 0x1010:
        raise WormErrorMaxRegisteredClientsReached()
    if rescode == 0x1011:
        raise WormErrorClientNotRegistered()
    if rescode == 0x1012:
        raise WormErrorExportUnacknowledgedData()
    if rescode == 0x1013:
        raise WormErrorClientHasUnfinishedTransactions()
    if rescode == 0x1014:
        raise WormErrorTseHasUnfinishedTransactions()
    if rescode == 0x1015:
        raise WormErrorTseNoResponseToFetch()
    if rescode == 0x1016:
        raise WormErrorNotAllowedExportInProgress()
    if rescode == 0x1017:
        raise WormErrorStoreFull()
    if rescode == 0x1050:
        raise WormErrorWrongStateNeedsPukChange()
    if rescode == 0x1051:
        raise WormErrorWrongStateNeedsPinChange()
    if rescode == 0x1053:
        raise WormErrorWrongStateNeedsActiveCtss()
    if rescode == 0x1054:
        raise WormErrorWrongStateNeedsSelfTest()
    if rescode == 0x1055:
        raise WormErrorWrongStateNeedsSelfTestPassed()
    if rescode == 0x1061:
        raise WormErrorFwuIntegrityFailure()
    if rescode == 0x1062:
        raise WormErrorFwuDecryptionFailure()
    if rescode == 0x1064:
        raise WormErrorFwuWrongFormat()
    if rescode == 0x1065:
        raise WormErrorFwuInternalError()
    if rescode == 0x1067:
        raise WormErrorFwuDowngradeProhibited()
    if rescode == 0x10FD:
        raise WormErrorTseAlreadyInitialized()
    if rescode == 0x10FE:
        raise WormErrorTseDecommissioned()
    if rescode == 0x10FF:
        raise WormErrorTseNotInitialized()
    if rescode == 0x1100:
        raise WormErrorAuthenticationFailed()
    if rescode == 0x1201:
        raise WormErrorAuthenticationPinBlocked()
    if rescode == 0x1202:
        raise WormErrorAuthenticationUserNotLoggedIn()
    if rescode == 0x1300:
        raise WormErrorSelfTestFailedFW()
    if rescode == 0x1310:
        raise WormErrorSelfTestFailedCSP()
    if rescode == 0x1320:
        raise WormErrorSelfTestFailedRNG()
    if rescode == 0x1400:
        raise WormErrorFwuBaseFWError()
    if rescode == 0x1500:
        raise WormErrorFwuFwextError()
    if rescode == 0x1600:
        raise WormErrorFwuCspError()
    if rescode == 0x2001:
        raise WormErrorExportNoneInProgress()
    if rescode == 0x2002:
        raise WormErrorExportRetry()
    if rescode == 0x2003:
        raise WormErrorExportNoDataAvailable()
    if rescode == 0xF000:
        raise WormErrorCmdNotFound()
    if rescode == 0xFF00:
        raise WormErrorSigError()
    if rescode == 0xFFFF:
        raise WormErrorFromCardLast()
    raise WormError(f"Unknown error ID {rescode}")



import logging
import time
from functools import wraps


def log_execution(f):
    logger = logging.getLogger(f.__module__)

    @wraps(f)
    def wrapper(*args, **kwargs):
        logger.debug(f"{f.__name__} called")
        t = time.time()
        try:
            r = f(*args, **kwargs)
        except:
            logger.debug(f"{f.__name__} failed in {time.time() - t}s")
            raise
        else:
            logger.debug(f"{f.__name__} returned in {time.time() - t}s")
        return r
    return wrapper